==> jfrog-annotation-processing-3.10.2-sources.jar: org/jfrog/common/config/diff/DiffUtils.java <==
package org.jfrog.common.config.diff;

import com.google.common.eventbus.EventBus;
import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.Validate;
import org.apache.commons.lang3.builder.Diff;
import org.apache.commons.lang3.builder.DiffBuilder;
import org.apache.commons.lang3.builder.DiffResult;
import org.apache.commons.lang3.builder.ToStringStyle;
import org.apache.commons.lang3.text.WordUtils;
import org.apache.commons.lang3.tuple.Pair;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * A utility class for common diff related operations
 *
 * @author Noam Shemesh
 */
public abstract class DiffUtils {

    public static final String NEW_MARKER = "$$new";
    public static final String DELETED_MARKER = "$$deleted";
    public static final String DELIMITER = "\u00BB";

    private DiffUtils() {}

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static <T> DiffResult diffMap(
            DiffFunctions diffFunctions, Class<T> c, Map<String, ? extends T> lhs, Map<String, ? extends T> rhs) {
        Map<String, ? extends T> localRhs = rhs != null ? rhs : Collections.emptyMap();
        Map<String, ? extends T> localLhs = lhs != null ? lhs : Collections.emptyMap();

        DiffBuilder db = new DiffBuilder(localLhs, localRhs, ToStringStyle.DEFAULT_STYLE, false);

        localLhs.forEach((key, lhsValue) -> diffByType(diffFunctions, c, localRhs, db, key, lhsValue));

        localRhs.keySet().stream()
                .filter(k -> !localLhs.containsKey(k))
                .forEach(key -> diffByType(diffFunctions, c, localRhs, db, key, null));

        return db.build();
    }

    private static <T> void diffByType(DiffFunctions diffFunctions, Class<T> c, Map<String, ? extends T> localRhs, DiffBuilder db,
            String key, T lhs) {
        if (diffFunctions.containsClass(c)) {
            appendDiffResult(db, key, diffInternalClass(diffFunctions, c, lhs, localRhs.get(key)));
        } else {
            db.append(key, lhs, localRhs.get(key));
        }
    }

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static <T> DiffResult diffCollection(
            DiffFunctions diffFunctions,
            Class<T> c,
            Function<T, Object> keyExtractor,
            Collection<? extends T> lhs, Collection<? extends T> rhs) {
        Collection<? extends T> localLhs = lhs != null ? lhs : Collections.emptyList();
        Collection<? extends T> localRhs = rhs != null ? rhs : Collections.emptyList();
        Function<T, String> keyToString = t -> "" + keyExtractor.apply(t);
        return diffMap(diffFunctions, c,
                localLhs.stream().collect(Collectors.toMap(keyToString, Function.identity())),
                localRhs.stream().collect(Collectors.toMap(keyToString, Function.identity()))
        );
    }

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static <T> DiffResult diffInternalClass(
            DiffFunctions diffFunctions,
            Class<T> c,
            T lhs, T rhs) {
        if (lhs == null || rhs == null) {
            return new DiffBuilder(emptyOrNotEmpty(lhs), emptyOrNotEmpty(rhs), ToStringStyle.DEFAULT_STYLE, false)
                    .append(lhs == null ? NEW_MARKER : DELETED_MARKER, lhs, rhs)
                    .build();
        }

        return diffFunctions.diffFor(c, lhs, rhs);
    }

    /**
     * That is the question (sorry.)
     */
    private static Optional<Object> emptyOrNotEmpty(Object object) {
        return Optional.ofNullable(object);
    }

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static DiffBuilder appendDiffResult(DiffBuilder diffBuilder, String fieldName,
            DiffResult diffResult) {
        validateFieldNameNotNull(fieldName);
        Validate.isTrue(diffResult != null, "Diff result cannot be null");

        for (Diff<?> diff : diffResult.getDiffs()) {
            diffBuilder.append(getDiffKey(fieldName, diff.getFieldName()),
                    diff.getLeft(), diff.getRight());
        }

        return diffBuilder;
    }

    public static <T> DiffResult diffCasted(
            DiffFunctions diffFunctions,
            Class<T> castTo,
            Pair<? super T, ? super T> toCompare) {
        if ((toCompare.getLeft() != null && !castTo.isInstance(toCompare.getLeft())) ||
                toCompare.getRight() != null && !castTo.isInstance(toCompare.getRight())) {
            throw new IllegalStateException("Couldn't cast object to " + castTo);
        }

        return diffInternalClass(diffFunctions, castTo,
                castTo.cast(toCompare.getLeft()), castTo.cast(toCompare.getRight()));
    }

    public static <T> Map<String, Set<T>> diffForField(Collection<T> diffs, Function<T, String> getFieldName) {
        return hierarchicalPaths(diffs, getFieldName);
    }

    private static String getDiffKey(String fieldName, String key) {
        return fieldName + DELIMITER + key;
    }

    private static void validateFieldNameNotNull(String fieldName) {
        Validate.isTrue(fieldName != null, "Field name cannot be null");
    }

    public static <T> void notifyHierarchicalEventBusIfPresent(
            Map<String, EventBus> eventBuses,
            DiffResult changes,
            Function<Collection<Diff<?>>, T> computeDiffResult
    ) {
        hierarchicalPaths(changes.getDiffs(), Diff::getFieldName)
                .forEach((key, diffList) -> Optional.ofNullable(eventBuses.get(key))
                        .ifPresent(value -> value.post(computeDiffResult.apply(diffList)))
                );
    }

    /**
     * Returns the fieldName and all of its parents, for example:
     * for <pre>grandparent.parent.child</pre> the result will be:
     *
     * key                                  value
     * <pre>grandparent</pre>               <pre>grandparent.parent.child</pre>
     * <pre>grandparent.parent</pre>        <pre>grandparent.parent.child</pre>
     * <pre>grandparent.parent.child</pre>  <pre>grandparent.parent.child</pre>
     */
    public static <T> Map<String, Set<T>> hierarchicalPaths(Collection<T> changes, Function<T, String> getFieldName) {
        return changes
                .stream()
                .flatMap(diff -> extractAllParents(diff, getFieldName))
                .reduce((Map<String, Set<T>>) new HashMap<String, Set<T>>(), DiffUtils::appendPath, DiffUtils::combinePaths);
    }

    private static <T> Map<String, Set<T>> appendPath(Map<String, Set<T>> allPaths, Map.Entry<String, T> currPath) {
        allPaths.computeIfAbsent(currPath.getKey(), (a) -> new HashSet<>()).add(currPath.getValue());
        return allPaths;
    }

    private static <T> Map<String, Set<T>> combinePaths(Map<String, Set<T>> allPaths1, Map<String, Set<T>> allPaths2) {
        Map<String, Set<T>> accRes = new HashMap<>();
        accRes.putAll(allPaths1);
        accRes.putAll(allPaths2);
        return accRes;
    }

    private static <T> Stream<Map.Entry<String, T>> extractAllParents(T diff, Function<T, String> getFieldName) {
        Map<String, T> diffs = new HashMap<>();
        String fullName = getFieldName.apply(diff);
        diffs.put(fullName, diff);
        int i = 0;
        int prev = fullName.length();
        while ((i = fullName.lastIndexOf(DELIMITER, prev)) > 0) {
            diffs.put(fullName.substring(0, i), diff);
            prev = i - 1;
        }
        return diffs.entrySet().stream();
    }

    public static String toFieldName(String method) {
        String noGet = method.replaceFirst("^(get|is)", "");
        return StringUtils.uncapitalize(WordUtils.capitalize(noGet.replace(".", " ")).replace(" ", ""));
    }

    public static String toMethodName(String fieldName, boolean isPrefix) {
        return (isPrefix ? "is" : "get") + StringUtils.capitalize(fieldName);
    }

    public static Method fieldToMethod(Field field) {
        try {
            return field.getDeclaringClass().getMethod(DiffUtils.toMethodName(field.getName(), isBoolean(field)));
        } catch (NoSuchMethodException e) {
            throw new IllegalStateException(e);
        }
    }

    public static Field methodToField(Method method) {
        try {
            return method.getDeclaringClass().getDeclaredField(DiffUtils.toFieldName(method.getName()));
        } catch (NoSuchFieldException e) {
            throw new IllegalStateException(e);
        }
    }

    public static boolean isBoolean(Field field) {
        return Boolean.class.equals(ClassUtils.primitiveToWrapper(field.getType()));
    }

    public static boolean isLeafChanged(String fieldName, String fieldToFind) {
        return fieldName.endsWith(fieldToFind);
    }

}

==> jfrog-annotation-processing-4.9.0-sources.jar: org/jfrog/common/config/diff/DiffUtils.java <==
package org.jfrog.common.config.diff;

import com.google.common.eventbus.EventBus;
import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.Validate;
import org.apache.commons.lang3.builder.Diff;
import org.apache.commons.lang3.builder.DiffBuilder;
import org.apache.commons.lang3.builder.DiffResult;
import org.apache.commons.lang3.builder.ToStringStyle;
import org.apache.commons.lang3.text.WordUtils;
import org.apache.commons.lang3.tuple.Pair;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * A utility class for common diff related operations
 *
 * @author Noam Shemesh
 */
public abstract class DiffUtils {

    public static final String NEW_MARKER = "$$new";
    public static final String DELETED_MARKER = "$$deleted";
    public static final String DELIMITER = "\u00BB";

    private DiffUtils() {
    }

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static <T> DiffResult diffMap(
            DiffFunctions diffFunctions, Class<T> c, Map<String, ? extends T> lhs, Map<String, ? extends T> rhs) {
        Map<String, ? extends T> localRhs = rhs != null ? rhs : Collections.emptyMap();
        Map<String, ? extends T> localLhs = lhs != null ? lhs : Collections.emptyMap();

        DiffBuilder db = new DiffBuilder(localLhs, localRhs, ToStringStyle.DEFAULT_STYLE, false);

        localLhs.forEach((key, lhsValue) -> diffByType(diffFunctions, c, localRhs, db, key, lhsValue));

        localRhs.keySet().stream()
                .filter(k -> !localLhs.containsKey(k))
                .forEach(key -> diffByType(diffFunctions, c, localRhs, db, key, null));

        return db.build();
    }

    private static <T> void diffByType(DiffFunctions diffFunctions, Class<T> c, Map<String, ? extends T> localRhs,
            DiffBuilder db,
            String key, T lhs) {
        if (diffFunctions.containsClass(c)) {
            appendDiffResult(db, key, diffInternalClass(diffFunctions, c, lhs, localRhs.get(key)));
        } else {
            db.append(key, lhs, localRhs.get(key));
        }
    }

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static <T> DiffResult diffCollection(
            DiffFunctions diffFunctions,
            Class<T> c,
            Function<T, Object> keyExtractor,
            Collection<? extends T> lhs, Collection<? extends T> rhs) {
        Collection<? extends T> localLhs = lhs != null ? lhs : Collections.emptyList();
        Collection<? extends T> localRhs = rhs != null ? rhs : Collections.emptyList();
        Function<T, String> keyToString = t -> "" + keyExtractor.apply(t);
        return diffMap(diffFunctions, c,
                localLhs.stream().collect(Collectors.toMap(keyToString, Function.identity())),
                localRhs.stream().collect(Collectors.toMap(keyToString, Function.identity()))
        );
    }

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static <T> DiffResult diffInternalClass(
            DiffFunctions diffFunctions,
            Class<T> c,
            T lhs, T rhs) {
        if (lhs == null || rhs == null) {
            return new DiffBuilder(emptyOrNotEmpty(lhs), emptyOrNotEmpty(rhs), ToStringStyle.DEFAULT_STYLE, false)
                    .append(lhs == null ? NEW_MARKER : DELETED_MARKER, lhs, rhs)
                    .build();
        }

        return diffFunctions.diffFor(c, lhs, rhs);
    }

    /**
     * That is the question (sorry.)
     */
    private static Optional<Object> emptyOrNotEmpty(Object object) {
        return Optional.ofNullable(object);
    }

    /**
     * Static internal method for {@link DiffProcessor}, do not use outside context
     */
    public static DiffBuilder appendDiffResult(DiffBuilder diffBuilder, String fieldName,
            DiffResult diffResult) {
        validateFieldNameNotNull(fieldName);
        Validate.isTrue(diffResult != null, "Diff result cannot be null");

        for (Diff<?> diff : diffResult.getDiffs()) {
            diffBuilder.append(getDiffKey(fieldName, diff.getFieldName()),
                    diff.getLeft(), diff.getRight());
        }

        return diffBuilder;
    }

    public static <T> DiffResult diffCasted(
            DiffFunctions diffFunctions,
            Class<T> castTo,
            Pair<? super T, ? super T> toCompare) {
        if ((toCompare.getLeft() != null && !castTo.isInstance(toCompare.getLeft())) ||
                toCompare.getRight() != null && !castTo.isInstance(toCompare.getRight())) {
            throw new IllegalStateException("Couldn't cast object to " + castTo);
        }

        return diffInternalClass(diffFunctions, castTo,
                castTo.cast(toCompare.getLeft()), castTo.cast(toCompare.getRight()));
    }

    public static <T> Map<String, Set<T>> diffForField(Collection<T> diffs, Function<T, String> getFieldName) {
        return hierarchicalPaths(diffs, getFieldName);
    }

    private static String getDiffKey(String fieldName, String key) {
        return fieldName + DELIMITER + key;
    }

    private static void validateFieldNameNotNull(String fieldName) {
        Validate.isTrue(fieldName != null, "Field name cannot be null");
    }

    public static <T> void notifyHierarchicalEventBusIfPresent(
            Map<String, EventBus> eventBuses,
            DiffResult changes,
            Function<Collection<Diff<?>>, T> computeDiffResult
    ) {
        hierarchicalPaths(changes.getDiffs(), Diff::getFieldName)
                .forEach((key, diffList) -> Optional.ofNullable(eventBuses.get(key))
                        .ifPresent(value -> value.post(computeDiffResult.apply(diffList)))
                );
    }

    /**
     * Returns the fieldName and all of its parents, for example:
     * for <pre>grandparent.parent.child</pre> the result will be:
     *
     * key                                  value
     * <pre>grandparent</pre>               <pre>grandparent.parent.child</pre>
     * <pre>grandparent.parent</pre>        <pre>grandparent.parent.child</pre>
     * <pre>grandparent.parent.child</pre>  <pre>grandparent.parent.child</pre>
     */
    public static <T> Map<String, Set<T>> hierarchicalPaths(Collection<T> changes, Function<T, String> getFieldName) {
        return changes
                .stream()
                .flatMap(diff -> extractAllParents(diff, getFieldName))
                .reduce((Map<String, Set<T>>) new HashMap<String, Set<T>>(), DiffUtils::appendPath,
                        DiffUtils::combinePaths);
    }

    private static <T> Map<String, Set<T>> appendPath(Map<String, Set<T>> allPaths, Map.Entry<String, T> currPath) {
        allPaths.computeIfAbsent(currPath.getKey(), (a) -> new HashSet<>()).add(currPath.getValue());
        return allPaths;
    }

    private static <T> Map<String, Set<T>> combinePaths(Map<String, Set<T>> allPaths1, Map<String, Set<T>> allPaths2) {
        Map<String, Set<T>> accRes = new HashMap<>();
        accRes.putAll(allPaths1);
        accRes.putAll(allPaths2);
        return accRes;
    }

    private static <T> Stream<Map.Entry<String, T>> extractAllParents(T diff, Function<T, String> getFieldName) {
        Map<String, T> diffs = new HashMap<>();
        String fullName = getFieldName.apply(diff);
        diffs.put(fullName, diff);
        int i = 0;
        int prev = fullName.length();
        while ((i = fullName.lastIndexOf(DELIMITER, prev)) > 0) {
            diffs.put(fullName.substring(0, i), diff);
            prev = i - 1;
        }
        return diffs.entrySet().stream();
    }

    public static String toFieldName(String method) {
        String noGet = method.replaceFirst("^(get|is)", "");
        return StringUtils.uncapitalize(WordUtils.capitalize(noGet.replace(".", " ")).replace(" ", ""));
    }

    public static String toMethodName(String fieldName, boolean isPrefix) {
        return (isPrefix ? "is" : "get") + StringUtils.capitalize(fieldName);
    }

    public static Method fieldToMethod(Field field) {
        try {
            return field.getDeclaringClass().getMethod(DiffUtils.toMethodName(field.getName(), isBoolean(field)));
        } catch (NoSuchMethodException e) {
            throw new IllegalStateException(e);
        }
    }

    public static Field methodToField(Method method) {
        try {
            return method.getDeclaringClass().getDeclaredField(DiffUtils.toFieldName(method.getName()));
        } catch (NoSuchFieldException e) {
            throw new IllegalStateException(e);
        }
    }

    public static boolean isBoolean(Field field) {
        return Boolean.class.equals(ClassUtils.primitiveToWrapper(field.getType()));
    }

    public static boolean isLeafChanged(String fieldName, String fieldToFind) {
        return fieldName.endsWith(fieldToFind);
    }

}
